@@ -1,31 +1,29 @@
 # TODO: Break this module up into pieces. Maybe group by functionality tested
 # rather than the socket level-ness of it.
-from urllib3 import HTTPConnectionPool, HTTPSConnectionPool
-from urllib3.poolmanager import proxy_from_url
+from dummyserver.server import (
+    DEFAULT_CA,
+    DEFAULT_CERTS,
+    encrypt_key_pem,
+    get_unreachable_address,
+)
+from dummyserver.testcase import SocketDummyServerTestCase, consume_socket
+from urllib3 import HTTPConnectionPool, HTTPSConnectionPool, ProxyManager, util
+from urllib3._collections import HTTPHeaderDict
+from urllib3.connection import HTTPConnection, _get_default_user_agent
 from urllib3.exceptions import (
     MaxRetryError,
+    ProtocolError,
     ProxyError,
     ReadTimeoutError,
     SSLError,
-    ProtocolError,
 )
-from urllib3.response import httplib
-from urllib3 import util
-from urllib3.util import ssl_wrap_socket
-from urllib3.util import ssl_
-from urllib3.util.timeout import Timeout
+from urllib3.packages.six.moves import http_client as httplib
+from urllib3.poolmanager import proxy_from_url
+from urllib3.util import ssl_, ssl_wrap_socket
 from urllib3.util.retry import Retry
-from urllib3._collections import HTTPHeaderDict
+from urllib3.util.timeout import Timeout
 
-from dummyserver.testcase import SocketDummyServerTestCase, consume_socket
-from dummyserver.server import (
-    DEFAULT_CERTS,
-    DEFAULT_CA,
-    get_unreachable_address,
-    encrypt_key_pem,
-)
-
-from .. import onlyPy3, LogRecorder
+from .. import LogRecorder, has_alpn, onlyPy3
 
 try:
     from mimetools import Message as MimeToolMessage
@@ -35,28 +33,30 @@ except ImportError:
         pass
 
 
-from collections import OrderedDict
-import os.path
-from threading import Event
 import os
+import os.path
+import platform
 import select
-import socket
 import shutil
+import socket
 import ssl
+import sys
 import tempfile
-import mock
-
-import pytest
-import trustme
-
+from collections import OrderedDict
 from test import (
-    requires_ssl_context_keyfile_password,
-    SHORT_TIMEOUT,
     LONG_TIMEOUT,
+    SHORT_TIMEOUT,
     notPyPy2,
     notSecureTransport,
+    notWindows,
+    requires_ssl_context_keyfile_password,
     resolvesLocalhostFQDN,
 )
+from threading import Event
+
+import mock
+import pytest
+import trustme
 
 # Retry failed tests
 pytestmark = pytest.mark.flaky
@@ -101,7 +101,7 @@ class TestSNI(SocketDummyServerTestCase):
             sock.close()
 
         self._start_server(socket_handler)
-        with HTTPConnectionPool(self.host, self.port) as pool:
+        with HTTPSConnectionPool(self.host, self.port) as pool:
             try:
                 pool.request("GET", "/", retries=0)
             except MaxRetryError:  # We are violating the protocol
@@ -113,6 +113,35 @@ class TestSNI(SocketDummyServerTestCase):
             ), "missing hostname in SSL handshake"
 
 
+class TestALPN(SocketDummyServerTestCase):
+    def test_alpn_protocol_in_first_request_packet(self):
+        if not has_alpn():
+            pytest.skip("ALPN-support not available")
+
+        done_receiving = Event()
+        self.buf = b""
+
+        def socket_handler(listener):
+            sock = listener.accept()[0]
+
+            self.buf = sock.recv(65536)  # We only accept one packet
+            done_receiving.set()  # let the test know it can proceed
+            sock.close()
+
+        self._start_server(socket_handler)
+        with HTTPSConnectionPool(self.host, self.port) as pool:
+            try:
+                pool.request("GET", "/", retries=0)
+            except MaxRetryError:  # We are violating the protocol
+                pass
+            successful = done_receiving.wait(LONG_TIMEOUT)
+            assert successful, "Timed out waiting for connection accept"
+            for protocol in util.ALPN_PROTOCOLS:
+                assert (
+                    protocol.encode("ascii") in self.buf
+                ), "missing ALPN protocol in SSL handshake"
+
+
 class TestClientCerts(SocketDummyServerTestCase):
     """
     Tests for client certificate support.
@@ -420,7 +449,7 @@ class TestSocketClosing(SocketDummyServerTestCase):
 
             assert http.pool.qsize() == http.pool.maxsize
 
-    def test_read_timeout_dont_retry_method_not_in_whitelist(self):
+    def test_read_timeout_dont_retry_method_not_in_allowlist(self):
         timed_out = Event()
 
         def socket_handler(listener):
@@ -431,7 +460,7 @@ class TestSocketClosing(SocketDummyServerTestCase):
 
         self._start_server(socket_handler)
         with HTTPConnectionPool(
-            self.host, self.port, timeout=SHORT_TIMEOUT, retries=True
+            self.host, self.port, timeout=LONG_TIMEOUT, retries=True
         ) as pool:
             try:
                 with pytest.raises(ReadTimeoutError):
@@ -440,7 +469,7 @@ class TestSocketClosing(SocketDummyServerTestCase):
                 timed_out.set()
 
     def test_https_connection_read_timeout(self):
-        """ Handshake timeouts should fail with a Timeout"""
+        """Handshake timeouts should fail with a Timeout"""
         timed_out = Event()
 
         def socket_handler(listener):
@@ -453,7 +482,7 @@ class TestSocketClosing(SocketDummyServerTestCase):
 
         self._start_server(socket_handler)
         with HTTPSConnectionPool(
-            self.host, self.port, timeout=SHORT_TIMEOUT, retries=False
+            self.host, self.port, timeout=LONG_TIMEOUT, retries=False
         ) as pool:
             try:
                 with pytest.raises(ReadTimeoutError):
@@ -603,7 +632,7 @@ class TestSocketClosing(SocketDummyServerTestCase):
                 response.read()
 
     def test_retry_weird_http_version(self):
-        """ Retry class should handle httplib.BadStatusLine errors properly """
+        """Retry class should handle httplib.BadStatusLine errors properly"""
 
         def socket_handler(listener):
             sock = listener.accept()[0]
@@ -676,9 +705,8 @@ class TestSocketClosing(SocketDummyServerTestCase):
         self._start_server(socket_handler)
         with HTTPConnectionPool(self.host, self.port) as pool:
             poolsize = pool.pool.qsize()
-            timeout = Timeout(connect=LONG_TIMEOUT, read=SHORT_TIMEOUT)
             response = pool.urlopen(
-                "GET", "/", retries=0, preload_content=False, timeout=timeout
+                "GET", "/", retries=0, preload_content=False, timeout=LONG_TIMEOUT
             )
             try:
                 with pytest.raises(ReadTimeoutError):
@@ -748,7 +776,7 @@ class TestSocketClosing(SocketDummyServerTestCase):
             # leaking it. Because we don't want to hang this thread, we
             # actually use select.select to confirm that a new request is
             # coming in: this lets us time the thread out.
-            rlist, _, _ = select.select([listener], [], [], 1)
+            rlist, _, _ = select.select([listener], [], [])
             assert rlist
             new_sock = listener.accept()[0]
 
@@ -857,7 +885,7 @@ class TestSocketClosing(SocketDummyServerTestCase):
             # Expect a new request. Because we don't want to hang this thread,
             # we actually use select.select to confirm that a new request is
             # coming in: this lets us time the thread out.
-            rlist, _, _ = select.select([listener], [], [], 5)
+            rlist, _, _ = select.select([listener], [], [])
             assert rlist
             sock = listener.accept()[0]
             consume_socket(sock)
@@ -887,7 +915,7 @@ class TestSocketClosing(SocketDummyServerTestCase):
                 retries=1,
                 release_conn=False,
                 preload_content=False,
-                timeout=Timeout(connect=LONG_TIMEOUT, read=SHORT_TIMEOUT),
+                timeout=LONG_TIMEOUT,
             )
 
             # The connection should still be on the response object, and none
@@ -937,6 +965,7 @@ class TestProxyManager(SocketDummyServerTestCase):
                     b"Host: google.com",
                     b"Accept-Encoding: identity",
                     b"Accept: */*",
+                    b"User-Agent: " + _get_default_user_agent().encode("utf-8"),
                     b"",
                     b"",
                 ]
@@ -1133,6 +1162,33 @@ class TestProxyManager(SocketDummyServerTestCase):
             except MaxRetryError:
                 self.fail("Invalid IPv6 format in HTTP CONNECT request")
 
+    @pytest.mark.parametrize("target_scheme", ["http", "https"])
+    def test_https_proxymanager_connected_to_http_proxy(self, target_scheme):
+        if target_scheme == "https" and sys.version_info[0] == 2:
+            pytest.skip("HTTPS-in-HTTPS isn't supported on Python 2")
+
+        errored = Event()
+
+        def http_socket_handler(listener):
+            sock = listener.accept()[0]
+            sock.send(b"HTTP/1.0 501 Not Implemented\r\nConnection: close\r\n\r\n")
+            errored.wait()
+            sock.close()
+
+        self._start_server(http_socket_handler)
+        base_url = "https://%s:%d" % (self.host, self.port)
+
+        with ProxyManager(base_url, cert_reqs="NONE") as proxy:
+            with pytest.raises(MaxRetryError) as e:
+                proxy.request("GET", "%s://example.com" % target_scheme, retries=0)
+
+            errored.set()  # Avoid a ConnectionAbortedError on Windows.
+
+            assert type(e.value.reason) == ProxyError
+            assert "Your proxy appears to only use HTTP and not HTTPS" in str(
+                e.value.reason
+            )
+
 
 class TestSSL(SocketDummyServerTestCase):
     def test_ssl_failure_midway_through_conn(self):
@@ -1400,6 +1456,104 @@ class TestSSL(SocketDummyServerTestCase):
         with pytest.raises(SSLError):
             ssl_wrap_socket(None, ca_certs="/tmp/fake-file")
 
+    def test_ssl_custom_validation_failure_terminates(self, tmpdir):
+        """
+        Ensure that the underlying socket is terminated if custom validation fails.
+        """
+        server_closed = Event()
+
+        def is_closed_socket(sock):
+            try:
+                sock.settimeout(SHORT_TIMEOUT)  # Python 3
+                sock.recv(1)  # Python 2
+            except (OSError, socket.error):
+                return True
+            return False
+
+        def socket_handler(listener):
+            sock = listener.accept()[0]
+            try:
+                _ = ssl.wrap_socket(
+                    sock,
+                    server_side=True,
+                    keyfile=DEFAULT_CERTS["keyfile"],
+                    certfile=DEFAULT_CERTS["certfile"],
+                    ca_certs=DEFAULT_CA,
+                )
+            except ssl.SSLError as e:
+                assert "alert unknown ca" in str(e)
+                if is_closed_socket(sock):
+                    server_closed.set()
+
+        self._start_server(socket_handler)
+
+        # client uses a different ca
+        other_ca = trustme.CA()
+        other_ca_path = str(tmpdir / "ca.pem")
+        other_ca.cert_pem.write_to_path(other_ca_path)
+
+        with HTTPSConnectionPool(
+            self.host, self.port, cert_reqs="REQUIRED", ca_certs=other_ca_path
+        ) as pool:
+            with pytest.raises(SSLError):
+                pool.request("GET", "/", retries=False, timeout=LONG_TIMEOUT)
+        assert server_closed.wait(LONG_TIMEOUT), "The socket was not terminated"
+
+    # SecureTransport can read only small pieces of data at the moment.
+    # https://github.com/urllib3/urllib3/pull/2674
+    @notSecureTransport
+    @pytest.mark.skipif(
+        os.environ.get("CI") == "true" and platform.python_implementation() == "PyPy",
+        reason="too slow to run in CI",
+    )
+    @pytest.mark.parametrize(
+        "preload_content,read_amt", [(True, None), (False, None), (False, 2 ** 31)]
+    )
+    def test_requesting_large_resources_via_ssl(self, preload_content, read_amt):
+        """
+        Ensure that it is possible to read 2 GiB or more via an SSL
+        socket.
+        https://github.com/urllib3/urllib3/issues/2513
+        """
+        content_length = 2 ** 31  # (`int` max value in C) + 1.
+        ssl_ready = Event()
+
+        def socket_handler(listener):
+            sock = listener.accept()[0]
+            ssl_sock = ssl.wrap_socket(
+                sock,
+                server_side=True,
+                keyfile=DEFAULT_CERTS["keyfile"],
+                certfile=DEFAULT_CERTS["certfile"],
+                ca_certs=DEFAULT_CA,
+            )
+            ssl_ready.set()
+
+            while not ssl_sock.recv(65536).endswith(b"\r\n\r\n"):
+                continue
+
+            ssl_sock.send(
+                b"HTTP/1.1 200 OK\r\n"
+                b"Content-Type: text/plain\r\n"
+                b"Content-Length: %d\r\n\r\n" % content_length
+            )
+
+            chunks = 2
+            for i in range(chunks):
+                ssl_sock.sendall(b"0" * (content_length // chunks))
+
+            ssl_sock.close()
+            sock.close()
+
+        self._start_server(socket_handler)
+        ssl_ready.wait(5)
+        with HTTPSConnectionPool(
+            self.host, self.port, ca_certs=DEFAULT_CA, retries=False
+        ) as pool:
+            response = pool.request("GET", "/", preload_content=preload_content)
+            data = response.data if preload_content else response.read(read_amt)
+            assert len(data) == content_length
+
 
 class TestErrorWrapping(SocketDummyServerTestCase):
     def test_bad_statusline(self):
@@ -1433,9 +1587,9 @@ class TestHeaders(SocketDummyServerTestCase):
             r = pool.request("GET", "/")
             assert HEADERS == dict(r.headers.items())  # to preserve case sensitivity
 
-    def test_headers_are_sent_with_the_original_case(self):
-        headers = {"foo": "bar", "bAz": "quux"}
-        parsed_headers = {}
+    def start_parsing_handler(self):
+        self.parsed_headers = OrderedDict()
+        self.received_headers = []
 
         def socket_handler(listener):
             sock = listener.accept()[0]
@@ -1444,11 +1598,13 @@ class TestHeaders(SocketDummyServerTestCase):
             while not buf.endswith(b"\r\n\r\n"):
                 buf += sock.recv(65536)
 
-            headers_list = [header for header in buf.split(b"\r\n")[1:] if header]
+            self.received_headers = [
+                header for header in buf.split(b"\r\n")[1:] if header
+            ]
 
-            for header in headers_list:
+            for header in self.received_headers:
                 (key, value) = header.split(b": ")
-                parsed_headers[key.decode("ascii")] = value.decode("ascii")
+                self.parsed_headers[key.decode("ascii")] = value.decode("ascii")
 
             sock.send(
                 ("HTTP/1.1 204 No Content\r\nContent-Length: 0\r\n\r\n").encode("utf-8")
@@ -1457,6 +1613,26 @@ class TestHeaders(SocketDummyServerTestCase):
             sock.close()
 
         self._start_server(socket_handler)
+
+    def test_headers_are_sent_with_the_original_case(self):
+        headers = {"foo": "bar", "bAz": "quux"}
+
+        self.start_parsing_handler()
+        expected_headers = {
+            "Accept-Encoding": "identity",
+            "Host": "{0}:{1}".format(self.host, self.port),
+            "User-Agent": _get_default_user_agent(),
+        }
+        expected_headers.update(headers)
+
+        with HTTPConnectionPool(self.host, self.port, retries=False) as pool:
+            pool.request("GET", "/", headers=HTTPHeaderDict(headers))
+            assert expected_headers == self.parsed_headers
+
+    def test_ua_header_can_be_overridden(self):
+        headers = {"uSeR-AgENt": "Definitely not urllib3!"}
+
+        self.start_parsing_handler()
         expected_headers = {
             "Accept-Encoding": "identity",
             "Host": "{0}:{1}".format(self.host, self.port),
@@ -1465,7 +1641,7 @@ class TestHeaders(SocketDummyServerTestCase):
 
         with HTTPConnectionPool(self.host, self.port, retries=False) as pool:
             pool.request("GET", "/", headers=HTTPHeaderDict(headers))
-            assert expected_headers == parsed_headers
+            assert expected_headers == self.parsed_headers
 
     def test_request_headers_are_sent_in_the_original_order(self):
         # NOTE: Probability this test gives a false negative is 1/(K!)
@@ -1477,69 +1653,26 @@ class TestHeaders(SocketDummyServerTestCase):
             (u"X-Header-%d" % i, str(i)) for i in reversed(range(K))
         ]
 
-        actual_request_headers = []
+        def filter_non_x_headers(d):
+            return [(k, v) for (k, v) in d.items() if k.startswith("X-Header-")]
 
-        def socket_handler(listener):
-            sock = listener.accept()[0]
+        request_headers = OrderedDict()
 
-            buf = b""
-            while not buf.endswith(b"\r\n\r\n"):
-                buf += sock.recv(65536)
-
-            headers_list = [header for header in buf.split(b"\r\n")[1:] if header]
-
-            for header in headers_list:
-                (key, value) = header.split(b": ")
-                if not key.decode("ascii").startswith(u"X-Header-"):
-                    continue
-                actual_request_headers.append(
-                    (key.decode("ascii"), value.decode("ascii"))
-                )
-
-            sock.send(
-                (u"HTTP/1.1 204 No Content\r\nContent-Length: 0\r\n\r\n").encode(
-                    "ascii"
-                )
-            )
-
-            sock.close()
-
-        self._start_server(socket_handler)
+        self.start_parsing_handler()
 
         with HTTPConnectionPool(self.host, self.port, retries=False) as pool:
             pool.request("GET", "/", headers=OrderedDict(expected_request_headers))
-            assert expected_request_headers == actual_request_headers
+            request_headers = filter_non_x_headers(self.parsed_headers)
+            assert expected_request_headers == request_headers
 
     @resolvesLocalhostFQDN
     def test_request_host_header_ignores_fqdn_dot(self):
-
-        received_headers = []
-
-        def socket_handler(listener):
-            sock = listener.accept()[0]
-
-            buf = b""
-            while not buf.endswith(b"\r\n\r\n"):
-                buf += sock.recv(65536)
-
-            for header in buf.split(b"\r\n")[1:]:
-                if header:
-                    received_headers.append(header)
-
-            sock.send(
-                (u"HTTP/1.1 204 No Content\r\nContent-Length: 0\r\n\r\n").encode(
-                    "ascii"
-                )
-            )
-
-            sock.close()
-
-        self._start_server(socket_handler)
+        self.start_parsing_handler()
 
         with HTTPConnectionPool(self.host + ".", self.port, retries=False) as pool:
             pool.request("GET", "/")
             self.assert_header_received(
-                received_headers, "Host", "%s:%s" % (self.host, self.port)
+                self.received_headers, "Host", "%s:%s" % (self.host, self.port)
             )
 
     def test_response_headers_are_returned_in_the_original_order(self):
@@ -1798,3 +1931,83 @@ class TestRetryPoolSizeDrainFail(SocketDummyServerTestCase):
         ) as pool:
             pool.urlopen("GET", "/not_found", preload_content=False)
             assert pool.num_connections == 1
+
+
+class TestBrokenPipe(SocketDummyServerTestCase):
+    @notWindows
+    def test_ignore_broken_pipe_errors(self, monkeypatch):
+        # On Windows an aborted connection raises an error on
+        # attempts to read data out of a socket that's been closed.
+        sock_shut = Event()
+        orig_connect = HTTPConnection.connect
+        # a buffer that will cause two sendall calls
+        buf = "a" * 1024 * 1024 * 4
+
+        def connect_and_wait(*args, **kw):
+            ret = orig_connect(*args, **kw)
+            assert sock_shut.wait(5)
+            return ret
+
+        def socket_handler(listener):
+            for i in range(2):
+                sock = listener.accept()[0]
+                sock.send(
+                    b"HTTP/1.1 404 Not Found\r\n"
+                    b"Connection: close\r\n"
+                    b"Content-Length: 10\r\n"
+                    b"\r\n"
+                    b"xxxxxxxxxx"
+                )
+                sock.shutdown(socket.SHUT_RDWR)
+                sock_shut.set()
+                sock.close()
+
+        monkeypatch.setattr(HTTPConnection, "connect", connect_and_wait)
+        self._start_server(socket_handler)
+        with HTTPConnectionPool(self.host, self.port) as pool:
+            r = pool.request("POST", "/", body=buf)
+            assert r.status == 404
+            assert r.headers["content-length"] == "10"
+            assert r.data == b"xxxxxxxxxx"
+
+            r = pool.request("POST", "/admin", chunked=True, body=buf)
+            assert r.status == 404
+            assert r.headers["content-length"] == "10"
+            assert r.data == b"xxxxxxxxxx"
+
+
+class TestMultipartResponse(SocketDummyServerTestCase):
+    def test_multipart_assert_header_parsing_no_defects(self):
+        def socket_handler(listener):
+            for _ in range(2):
+                sock = listener.accept()[0]
+                while not sock.recv(65536).endswith(b"\r\n\r\n"):
+                    pass
+
+                sock.sendall(
+                    b"HTTP/1.1 404 Not Found\r\n"
+                    b"Server: example.com\r\n"
+                    b"Content-Type: multipart/mixed; boundary=36eeb8c4e26d842a\r\n"
+                    b"Content-Length: 73\r\n"
+                    b"\r\n"
+                    b"--36eeb8c4e26d842a\r\n"
+                    b"Content-Type: text/plain\r\n"
+                    b"\r\n"
+                    b"1\r\n"
+                    b"--36eeb8c4e26d842a--\r\n",
+                )
+                sock.close()
+
+        self._start_server(socket_handler)
+        from urllib3.connectionpool import log
+
+        with mock.patch.object(log, "warning") as log_warning:
+            with HTTPConnectionPool(self.host, self.port, timeout=3) as pool:
+                resp = pool.urlopen("GET", "/")
+                assert resp.status == 404
+                assert (
+                    resp.headers["content-type"]
+                    == "multipart/mixed; boundary=36eeb8c4e26d842a"
+                )
+                assert len(resp.data) == 73
+                log_warning.assert_not_called()
